/*****************************************************************************
 * auhal.c: AUHAL and Coreaudio output plugin
 *****************************************************************************
 * Copyright (C) 2005 the VideoLAN team, 2007 Jesse Chappell
 * $Id$
 *
 * Authors: Derk-Jan Hartman <hartman at videolan dot org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

/*****************************************************************************
 * Preamble
 *****************************************************************************/
#include <cstring>
#include <cstdlib>
#include <iostream>

#include <unistd.h>
#include <pthread.h>

#include <sys/time.h>

#include "CoreAudioOutput.hpp"

#include <CoreAudio/CoreAudio.h>
#include <AudioUnit/AudioUnitProperties.h>
#include <AudioUnit/AudioUnitParameters.h>
#include <AudioUnit/AudioOutputUnit.h>
#include <AudioToolbox/AudioFormat.h>

using namespace AC3JACK::COREAUDIO;
using namespace std;

#define STREAM_FORMAT_MSG( pre, sfm ) \
    pre "[%ld][%4.4s][%ld][%ld][%ld][%ld][%ld][%ld]", \
    (UInt32)sfm.mSampleRate, (char *)&sfm.mFormatID, \
    sfm.mFormatFlags, sfm.mBytesPerPacket, \
    sfm.mFramesPerPacket, sfm.mBytesPerFrame, \
    sfm.mChannelsPerFrame, sfm.mBitsPerChannel

#define STREAM_FORMAT_MSG_FULL( pre, sfm ) \
    pre ":\nsamplerate: [%ld]\nFormatID: [%4.4s]\nFormatFlags: [%ld]\nBypesPerPacket: [%ld]\nFramesPerPacket: [%ld]\nBytesPerFrame: [%ld]\nChannelsPerFrame: [%ld]\nBitsPerChannel[%ld]", \
    (UInt32)sfm.mSampleRate, (char *)&sfm.mFormatID, \
    sfm.mFormatFlags, sfm.mBytesPerPacket, \
    sfm.mFramesPerPacket, sfm.mBytesPerFrame, \
    sfm.mChannelsPerFrame, sfm.mBitsPerChannel

#define BUFSIZE 0xffffff
#define AOUT_VAR_SPDIF_FLAG 0xf00000

#define VLC_TRUE 1
#define VLC_FALSE 0

/*
 * TODO:
 * - clean up the debug info
 * - clean up C99'isms
 * - be better at changing stream setup or devices setup changes while playing.
 * - fix 6.1 and 7.1
 */


CoreAudioOutput::CoreAudioOutput(DataSource * source, double samplerate)
	: _samplerate(samplerate), _source(source) 
{
	/* Allocate structure */
	_sys = new aout_sys_t;
	_sys->i_stream_index = -1;
	_sys->i_selected_dev = 0;
	_allow_pcm = false;

	_opened = false;
}

CoreAudioOutput::~CoreAudioOutput()
{
	if (_sys && _opened) {
		Close();
	}
}

void CoreAudioOutput::set_device_stream (const std::string & idinfo)
{
	_devstr = idinfo;
	// TODO, parse it
	int devid=-1, streamindex = -1;
	int ret;
	bool wasopen = false;

	if (_opened) {
		wasopen = true;
		Close();
	}

	ret = sscanf(idinfo.c_str(), "%d:%d", &devid, &streamindex);
	if (ret > 0) {

		_sys->i_selected_dev = devid;
		cerr << "setting devdi to: " << devid << endl;
		if (ret == 2) {
			_sys->i_stream_index = streamindex;
			cerr << "setting streami to: " << streamindex << endl;
		}

		if (wasopen) {
			Open();
		}
	}
	
	cerr << "after device stream CLOSE" << endl;
}

/*****************************************************************************
 * Open: open macosx audio output
 *****************************************************************************/
int CoreAudioOutput::Open()
{
    OSStatus                err = noErr;
    UInt32                  i_param_size = 0;
    aout_sys_t       *p_sys = NULL;
    int                     b_alive = VLC_FALSE;

    if( _sys == NULL )
    {
        return( -2 );
    }

    p_sys = _sys;
    p_sys->i_default_dev = 0;
    //p_sys->i_selected_dev = 0;
    p_sys->i_devices = 0;
    p_sys->b_supports_digital = VLC_FALSE;
    p_sys->b_digital = VLC_FALSE;
    p_sys->au_component = NULL;
    p_sys->au_unit = NULL;
    p_sys->clock_diff =  0;
    p_sys->i_read_bytes = 0;
    p_sys->i_total_bytes = 0;
    p_sys->i_hog_pid = -1;
    p_sys->i_stream_id = 0;
    //p_sys->i_stream_index = -1;
    p_sys->b_revert = VLC_FALSE;
    p_sys->b_changed_mixing = VLC_FALSE;
    //memset( p_sys->p_remainder_buffer, 0, sizeof(uint8_t) * BUFSIZE );

    
    /* Build a list of devices */
    if (_device_list.empty()) {
	    Probe();
    }

    /* What device do we want? */
    // TODO choose device

    p_sys->b_supports_digital = ( p_sys->i_selected_dev & AOUT_VAR_SPDIF_FLAG ) ? VLC_TRUE : VLC_FALSE;
    p_sys->i_selected_dev = p_sys->i_selected_dev  & ~AOUT_VAR_SPDIF_FLAG ; /* remove SPDIF flag to get the true DeviceID */


    /* Check if the desired device is alive and usable */
    /* TODO: add a callback to the device to alert us if the device dies */
    i_param_size = sizeof( b_alive );
    err = AudioDeviceGetProperty( p_sys->i_selected_dev, 0, FALSE,
                                  kAudioDevicePropertyDeviceIsAlive,
                                  &i_param_size, &b_alive );

    if( err != noErr )
    {
	    fprintf(stderr, "could not check whether device is alive: %4.4s\n", (char *)&err );
	    goto error;
    }
    
    if( b_alive == VLC_FALSE )
    {
        fprintf(stderr, "selected audio device is not alive, switching to default device\n" ); 
        p_sys->i_selected_dev = p_sys->i_default_dev;
    }

    i_param_size = sizeof( p_sys->i_hog_pid );
    err = AudioDeviceGetProperty( p_sys->i_selected_dev, 0, FALSE,
                                  kAudioDevicePropertyHogMode,
                                  &i_param_size, &p_sys->i_hog_pid );

    if( err != noErr )
    {
        /* This is not a fatal error. Some drivers simply don't support this property */
        fprintf(stderr, "could not check whether device is hogged: %4.4s\n",
                 (char *)&err );
        p_sys->i_hog_pid = -1;
    }

    if( p_sys->i_hog_pid != -1 && p_sys->i_hog_pid != getpid() )
    {
        fprintf(stderr, "Selected audio device is exclusively in use by another program.\n" );
        goto error;
    }

    /* Check for Digital mode or Analog output mode */
    if( p_sys->b_supports_digital )
    {
	    if( OpenSPDIF() ) {
		    _opened = true;
		    return 1;
	    }
    }
    else
    {
	    if( OpenAnalog() ) {
		    _opened = true;
		    return 1;
	    }
    }
    
    

error:
   
    //if( _sys )  {
//	    delete _sys;
//	    _sys = 0;
    //  }
    return 0;
}


/*****************************************************************************
 * Open: open and setup a HAL AudioUnit to do analog (multichannel) audio output
 *****************************************************************************/

int CoreAudioOutput::OpenAnalog()
{

    aout_sys_t           *p_sys = _sys;
    OSStatus                    err = noErr;
    UInt32                      i_param_size = 0;
    int                         i_original; 
    int                         i;
    ComponentDescription        desc;
    AudioStreamBasicDescription DeviceFormat;
    AudioChannelLayout          *layout;
    AudioChannelLayout          new_layout;
    AURenderCallbackStruct      input;

    /* Lets go find our Component */
    desc.componentType = kAudioUnitType_Output;
    desc.componentSubType = kAudioUnitSubType_HALOutput;
    desc.componentManufacturer = kAudioUnitManufacturer_Apple;
    desc.componentFlags = 0;
    desc.componentFlagsMask = 0;

    p_sys->au_component = FindNextComponent( NULL, &desc );
    if( p_sys->au_component == NULL )
    {
        fprintf(stderr, "we cannot find our HAL component\n" );
        return VLC_FALSE;
    }

    err = OpenAComponent( p_sys->au_component, &p_sys->au_unit );
    if( err )
    {
        fprintf(stderr, "we cannot open our HAL component\n" );
        return VLC_FALSE;
    }
    
    /* AU initiliaze */
    verify_noerr( AudioUnitInitialize(p_sys->au_unit) );
    
    // enable i/o 
    UInt32 enableIO = 1;
    verify_noerr(AudioUnitSetProperty(p_sys->au_unit, kAudioOutputUnitProperty_EnableIO, kAudioUnitScope_Output, 0, &enableIO, sizeof(enableIO)));

    /* Set the device we will use for this output unit */
    verify_noerr( AudioUnitSetProperty( p_sys->au_unit,
                         kAudioOutputUnitProperty_CurrentDevice,
                         kAudioUnitScope_Global,
                         0,
                         &p_sys->i_selected_dev,
                         sizeof( AudioDeviceID )));
                         
    /* Get the current format */
    i_param_size = sizeof(AudioStreamBasicDescription);

    verify_noerr( AudioUnitGetProperty( p_sys->au_unit,
                                   kAudioUnitProperty_StreamFormat,
                                   kAudioUnitScope_Input,
                                   0,
                                   &DeviceFormat,
                                   &i_param_size ));
                                   
    fprintf(stderr, STREAM_FORMAT_MSG( "current format is: ", DeviceFormat ) );
    fprintf(stderr, "\n");

#if 0
    /* Get the channel layout of the device side of the unit (vlc -> unit -> device) */
    verify_noerr( AudioUnitGetPropertyInfo( p_sys->au_unit,
                                   kAudioDevicePropertyPreferredChannelLayout,
                                   kAudioUnitScope_Output,
                                   0,
                                   &i_param_size,
                                   NULL ));

    layout = (AudioChannelLayout *)malloc( i_param_size);

    verify_noerr( AudioUnitGetProperty( p_sys->au_unit,
                                   kAudioDevicePropertyPreferredChannelLayout,
                                   kAudioUnitScope_Output,
                                   0,
                                   layout,
                                   &i_param_size ));
                                   
    /* We need to "fill out" the ChannelLayout, because there are multiple ways that it can be set */
    if( layout->mChannelLayoutTag == kAudioChannelLayoutTag_UseChannelBitmap)
    {
        /* bitmap defined channellayout */
        verify_noerr( AudioFormatGetProperty( kAudioFormatProperty_ChannelLayoutForBitmap,
                                sizeof( UInt32), &layout->mChannelBitmap,
                                &i_param_size,
                                layout ));
    }
    else if( layout->mChannelLayoutTag != kAudioChannelLayoutTag_UseChannelDescriptions )
    {
        /* layouttags defined channellayout */
        verify_noerr( AudioFormatGetProperty( kAudioFormatProperty_ChannelLayoutForTag,
                                sizeof( AudioChannelLayoutTag ), &layout->mChannelLayoutTag,
                                &i_param_size,
                                layout ));
    }

    fprintf(stderr, "layout of AUHAL has %d channels\n" , (int)layout->mNumberChannelDescriptions );
    
    /* Initialize the VLC core channel count */
    //p_aout->output.output.i_physical_channels = 0;
    //i_original = p_aout->output.output.i_original_channels & AOUT_CHAN_PHYSMASK;

    if( i_original == AOUT_CHAN_CENTER || layout->mNumberChannelDescriptions < 2 )
    {
        /* We only need Mono or cannot output more than 1 channel */
        p_aout->output.output.i_physical_channels = AOUT_CHAN_CENTER;
    }
    else if( i_original == (AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT) || layout->mNumberChannelDescriptions < 3 )
    {
        /* We only need Stereo or cannot output more than 2 channels */
        p_aout->output.output.i_physical_channels = AOUT_CHAN_RIGHT | AOUT_CHAN_LEFT;
    }
    else
    {
        /* We want more than stereo and we can do that */
        for( i = 0; i < layout->mNumberChannelDescriptions; i++ )
        {
            fprintf(stderr, "this is channel: %d\n", (int)layout->mChannelDescriptions[i].mChannelLabel );

            switch( layout->mChannelDescriptions[i].mChannelLabel )
            {
                case kAudioChannelLabel_Left:
                    p_aout->output.output.i_physical_channels |= AOUT_CHAN_LEFT;
                    continue;
                case kAudioChannelLabel_Right:
                    p_aout->output.output.i_physical_channels |= AOUT_CHAN_RIGHT;
                    continue;
                case kAudioChannelLabel_Center:
                    p_aout->output.output.i_physical_channels |= AOUT_CHAN_CENTER;
                    continue;
                case kAudioChannelLabel_LFEScreen:
                    p_aout->output.output.i_physical_channels |= AOUT_CHAN_LFE;
                    continue;
                case kAudioChannelLabel_LeftSurround:
                    p_aout->output.output.i_physical_channels |= AOUT_CHAN_REARLEFT;
                    continue;
                case kAudioChannelLabel_RightSurround:
                    p_aout->output.output.i_physical_channels |= AOUT_CHAN_REARRIGHT;
                    continue;
                case kAudioChannelLabel_RearSurroundLeft:
                    p_aout->output.output.i_physical_channels |= AOUT_CHAN_MIDDLELEFT;
                    continue;
                case kAudioChannelLabel_RearSurroundRight:
                    p_aout->output.output.i_physical_channels |= AOUT_CHAN_MIDDLERIGHT;
                    continue;
                case kAudioChannelLabel_CenterSurround:
                    p_aout->output.output.i_physical_channels |= AOUT_CHAN_REARCENTER;
                    continue;
                default:
                    fprintf(stderr, "Unrecognized channel form provided by driver: %d\n", (int)layout->mChannelDescriptions[i].mChannelLabel );
            }
        }
        if( p_aout->output.output.i_physical_channels == 0 )
        {
            p_aout->output.output.i_physical_channels = AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT;
            fprintf(stderr, "You should configure your speaker layout with Audio Midi Setup Utility in /Applications/Utilities. Now using Stereo mode.\n" );
        }
    }
    if( layout ) free( layout );

    fprintf(stderr, "selected %d physical channels for device output\n", aout_FormatNbChannels( &p_aout->output.output ) );
    fprintf(stderr, "VLC will output: %s\n", aout_FormatPrintChannels( &p_aout->output.output ));
#endif

    //memset (&new_layout, 0, sizeof(new_layout));
    //new_layout.mChannelLayoutTag = kAudioChannelLayoutTag_Stereo;

    int out_nChannels;
    err = get_total_channels(p_sys->i_selected_dev, &out_nChannels);
    if (err != noErr)   {
	    cerr << "error getting device output channel count" << endl;
    }
    SInt32 chanArr[out_nChannels];

    for (i = 0;	i < out_nChannels; i++) {
            chanArr[i] = -1;
    }


    chanArr[p_sys->i_stream_index] = 0;
    chanArr[p_sys->i_stream_index+1] = 1;
    cerr << "setting channel map to: " << chanArr[0] << "  " << chanArr[1] << "  totchan: " << out_nChannels << endl;

    err = AudioUnitSetProperty(p_sys->au_unit, kAudioOutputUnitProperty_ChannelMap, kAudioUnitScope_Output, 0, chanArr, sizeof(SInt32) * out_nChannels);
    if (err != noErr) {
	    cerr << "error setting channel map: " << err << endl;
    }

    /* Set up the format to be used */
    DeviceFormat.mSampleRate = _samplerate; // should use whatever it is
    DeviceFormat.mFormatID = kAudioFormatLinearPCM;

    /* We use 16 bit stereo, because that's what ac3 is producing */
    //DeviceFormat.mFormatFlags = kAudioFormatFlagsNativeFloatPacked;
    //DeviceFormat.mFormatFlags = kAudioFormatFlagIsSignedInteger|kAudioFormatFlagIsPacked|kAudioFormatFlagIsBigEndian;
    DeviceFormat.mFormatFlags = kAudioFormatFlagIsSignedInteger|kAudioFormatFlagIsPacked;
    DeviceFormat.mBitsPerChannel = 16;
    DeviceFormat.mChannelsPerFrame = 2;
    
    /* Calculate framesizes and stuff */
    DeviceFormat.mFramesPerPacket = 1;
    DeviceFormat.mBytesPerFrame = DeviceFormat.mBitsPerChannel * DeviceFormat.mChannelsPerFrame / 8;
    DeviceFormat.mBytesPerPacket = DeviceFormat.mBytesPerFrame * DeviceFormat.mFramesPerPacket;
 
    /* Set the desired format */
    i_param_size = sizeof(AudioStreamBasicDescription);
    verify_noerr( AudioUnitSetProperty( p_sys->au_unit,
                                   kAudioUnitProperty_StreamFormat,
                                   kAudioUnitScope_Input,
                                   0,
                                   &DeviceFormat,
                                   i_param_size ));
                                   
    fprintf(stderr, STREAM_FORMAT_MSG( "we set the AU format: " , DeviceFormat ) );
    fprintf(stderr, "\n");
    /* Retrieve actual format */
    verify_noerr( AudioUnitGetProperty( p_sys->au_unit,
                                   kAudioUnitProperty_StreamFormat,
                                   kAudioUnitScope_Input,
                                   0,
                                   &DeviceFormat,
                                   &i_param_size ));
                                   
    fprintf(stderr, STREAM_FORMAT_MSG( "the actual set AU format is " , DeviceFormat ) );
    fprintf(stderr,"\n");
   

    /* set the IOproc callback */
    input.inputProc = (AURenderCallback) RenderCallbackAnalog;
    input.inputProcRefCon = this;
    
    verify_noerr( AudioUnitSetProperty( p_sys->au_unit,
                            kAudioUnitProperty_SetRenderCallback,
                            kAudioUnitScope_Input,
                            0, &input, sizeof( input ) ) );

    input.inputProc = (AURenderCallback) RenderCallbackAnalog;
    input.inputProcRefCon = this;
    
    /* Set the new_layout as the layout VLC will use to feed the AU unit */
    //verify_noerr( AudioUnitSetProperty( p_sys->au_unit,
    //                       kAudioUnitProperty_AudioChannelLayout,
    //                       kAudioUnitScope_Input,
    //                       0, &new_layout, sizeof(new_layout) ) );
                            
    //if( new_layout.mNumberChannelDescriptions > 0 )
    //    free( new_layout.mChannelDescriptions );
    
    /* AU initiliaze */
//    verify_noerr( AudioUnitInitialize(p_sys->au_unit) );

    /* Find the difference between device clock and mdate clock */
    //p_sys->clock_diff = - (mtime_t)
    //    AudioConvertHostTimeToNanos( AudioGetCurrentHostTime() ) / 1000; 
    //p_sys->clock_diff += mdate();

    /* Start the AU */
    verify_noerr( AudioOutputUnitStart(p_sys->au_unit) );
    
    return VLC_TRUE;
}


/*****************************************************************************
 * Setup a encoded digital stream (SPDIF)
 *****************************************************************************/
int CoreAudioOutput::OpenSPDIF()
{
    aout_sys_t       *p_sys = _sys;
    OSStatus                err = noErr;
    UInt32                  i_param_size = 0, b_mix = 0;
    Boolean                 b_writeable = VLC_FALSE;
    AudioStreamID           *p_streams = NULL;
    int                     i = 0, i_streams = 0;

    /* Start doing the SPDIF setup proces */
    p_sys->b_digital = VLC_TRUE;

    /* Hog the device */
    i_param_size = sizeof( p_sys->i_hog_pid );
    p_sys->i_hog_pid = getpid() ;
    
    err = AudioDeviceSetProperty( p_sys->i_selected_dev, 0, 0, FALSE,
                                  kAudioDevicePropertyHogMode, i_param_size, &p_sys->i_hog_pid );
    
    if( err != noErr )
    {
        fprintf(stderr, "failed to set hogmode: [%4.4s]\n", (char *)&err );
        return VLC_FALSE;
    }

    /* Set mixable to false if we are allowed to */
    err = AudioDeviceGetPropertyInfo( p_sys->i_selected_dev, 0, FALSE, kAudioDevicePropertySupportsMixing,
                                    &i_param_size, &b_writeable );

    err = AudioDeviceGetProperty( p_sys->i_selected_dev, 0, FALSE, kAudioDevicePropertySupportsMixing,
                                    &i_param_size, &b_mix );
                                    
    if( !err && b_writeable )
    {
        b_mix = 0;
        err = AudioDeviceSetProperty( p_sys->i_selected_dev, 0, 0, FALSE,
                            kAudioDevicePropertySupportsMixing, i_param_size, &b_mix );
        p_sys->b_changed_mixing = VLC_TRUE;
    }
    
    if( err != noErr )
    {
        fprintf(stderr, "failed to set mixmode: [%4.4s]\n", (char *)&err );
        return VLC_FALSE;
    }

    /* Get a list of all the streams on this device */
    err = AudioDeviceGetPropertyInfo( p_sys->i_selected_dev, 0, FALSE,
                                      kAudioDevicePropertyStreams,
                                      &i_param_size, NULL );
    if( err != noErr )
    {
        fprintf(stderr, "could not get number of streams: [%4.4s]\n", (char *)&err );
        return VLC_FALSE;
    }
    
    i_streams = i_param_size / sizeof( AudioStreamID );
    p_streams = (AudioStreamID *)malloc( i_param_size );
    if( p_streams == NULL )
    {
        fprintf(stderr, "Out of memory\n" );
        return VLC_FALSE;
    }
    
    err = AudioDeviceGetProperty( p_sys->i_selected_dev, 0, FALSE,
                                    kAudioDevicePropertyStreams,
                                    &i_param_size, p_streams );
    
    if( err != noErr )
    {
        fprintf(stderr, "could not get number of streams: [%4.4s]\n", (char *)&err );
        if( p_streams ) free( p_streams );
        return VLC_FALSE;
    }

    for( i = 0; i < i_streams ; i++ )
    {
        /* Find a stream with a cac3 stream */
        AudioStreamBasicDescription *p_format_list = NULL;
        int                         i_formats = 0, j = 0;
        int                         b_usable = VLC_FALSE;
        
	if (p_sys->i_stream_index >= 0 && p_sys->i_stream_index != i) {
		continue;
	}

        /* Retrieve all the stream formats supported by each output stream */
        err = AudioStreamGetPropertyInfo( p_streams[i], 0,
                                          kAudioStreamPropertyPhysicalFormats,
                                          &i_param_size, NULL );
        if( err != noErr )
        {
            fprintf(stderr, "could not get number of streamformats: [%4.4s]\n", (char *)&err );
            continue;
        }
        
        i_formats = i_param_size / sizeof( AudioStreamBasicDescription );
        p_format_list = (AudioStreamBasicDescription *)malloc( i_param_size );
        if( p_format_list == NULL )
        {
            fprintf(stderr, "could not malloc the memory\n" );
            continue;
        }
        
        err = AudioStreamGetProperty( p_streams[i], 0,
                                          kAudioStreamPropertyPhysicalFormats,
                                          &i_param_size, p_format_list );
        if( err != noErr )
        {
            fprintf(stderr, "could not get the list of streamformats: [%4.4s]\n", (char *)&err );
            if( p_format_list) free( p_format_list);
            continue;
        }

        /* Check if one of the supported formats is a digital format */
        for( j = 0; j < i_formats; j++ )
        {
            if( p_format_list[j].mFormatID == 'IAC3' ||
                  p_format_list[j].mFormatID == kAudioFormat60958AC3 ||
		(_allow_pcm && p_format_list[j].mFormatID == kAudioFormatLinearPCM ))
            {
                b_usable = VLC_TRUE;
                break;
            }
        }
        
        if( b_usable)
        {
            /* if this stream supports a digital (cac3) format (or pcm disguise), then go set it. */
            int i_requested_rate_format = -1;
            int i_current_rate_format = -1;
            int i_backup_rate_format = -1;

            p_sys->i_stream_id = p_streams[i];
            p_sys->i_stream_index = i;

            if( p_sys->b_revert == VLC_FALSE )
            {
                /* Retrieve the original format of this stream first if not done so already */
                i_param_size = sizeof( p_sys->sfmt_revert );
                err = AudioStreamGetProperty( p_sys->i_stream_id, 0,
                                              kAudioStreamPropertyPhysicalFormat,
                                              &i_param_size, 
                                              &p_sys->sfmt_revert );
                if( err != noErr )
                {
                    fprintf(stderr, "could not retrieve the original streamformat: [%4.4s]\n", (char *)&err );
                    continue; 
                }
                p_sys->b_revert = VLC_TRUE;
            }

            for( j = 0; j < i_formats; j++ )
            {
                if( p_format_list[j].mFormatID == 'IAC3' ||
                      p_format_list[j].mFormatID == kAudioFormat60958AC3 ||
		    (_allow_pcm && p_format_list[j].mFormatID == kAudioFormatLinearPCM ))
                {
                    if( p_format_list[j].mSampleRate == _samplerate )
                    {
                        i_requested_rate_format = j;
                        break;
                    }
                    else if( p_format_list[j].mSampleRate == p_sys->sfmt_revert.mSampleRate )
                    {
                        i_current_rate_format = j;
                    }
                    else
                    {
                        if( i_backup_rate_format < 0 || p_format_list[j].mSampleRate > p_format_list[i_backup_rate_format].mSampleRate )
                            i_backup_rate_format = j;
                    }
                }
                    
            }
            
            if( i_requested_rate_format >= 0 ) /* We prefer to output at the samplerate of the original audio */
                p_sys->stream_format = p_format_list[i_requested_rate_format];
            else if( i_current_rate_format >= 0 ) /* If not possible, we will try to use the current samplerate of the device */
                p_sys->stream_format = p_format_list[i_current_rate_format];
            else p_sys->stream_format = p_format_list[i_backup_rate_format]; /* And if we have to, any digital format will be just fine (highest rate possible) */
        }
        if( p_format_list ) free( p_format_list );
    }
    if( p_streams ) free( p_streams );

    fprintf(stderr, STREAM_FORMAT_MSG( "original stream format: ", p_sys->sfmt_revert ) );
    fprintf(stderr, "\n");

    if( !AudioStreamChangeFormat(p_sys->i_stream_id, p_sys->stream_format ) )
        return VLC_FALSE;

    /* Set the format flags */
    if( p_sys->stream_format.mFormatFlags & kAudioFormatFlagIsBigEndian )
	    _hw_bigendian = true;
    else
	    _hw_bigendian = false;

    //p_aout->output.output.i_bytes_per_frame = AOUT_SPDIF_SIZE;
    //p_aout->output.output.i_frame_length = A52_FRAME_NB;
    //p_aout->output.i_nb_samples = p_aout->output.output.i_frame_length;
    _output_rate = (unsigned int)p_sys->stream_format.mSampleRate;
    //aout_FormatPrepare( &p_aout->output.output );
    //aout_VolumeNoneInit( p_aout );

    /* Add IOProc callback */
    err = AudioDeviceAddIOProc( p_sys->i_selected_dev,
                                (AudioDeviceIOProc)RenderCallbackSPDIF,
                                (void *)this);
    if( err != noErr )
    {
        fprintf(stderr, "AudioDeviceAddIOProc failed: [%4.4s]\n", (char *)&err );
        return VLC_FALSE;
    }

    /* Check for the difference between the Device clock and mdate */
    //p_sys->clock_diff = - (mtime_t)
    //    AudioConvertHostTimeToNanos( AudioGetCurrentHostTime() ) / 1000; 
    //p_sys->clock_diff += mdate();
 
    /* Start device */
    err = AudioDeviceStart( p_sys->i_selected_dev, (AudioDeviceIOProc)RenderCallbackSPDIF ); 
    if( err != noErr )
    {
        fprintf(stderr, "AudioDeviceStart failed: [%4.4s]\n", (char *)&err );

        err = AudioDeviceRemoveIOProc( p_sys->i_selected_dev, 
                                       (AudioDeviceIOProc)RenderCallbackSPDIF );
        if( err != noErr )
        {
            fprintf(stderr, "AudioDeviceRemoveIOProc failed: [%4.4s]\n", (char *)&err );
        }
        return VLC_FALSE;
    }

    return VLC_TRUE;
}


/*****************************************************************************
 * Close: Close HAL AudioUnit
 *****************************************************************************/
void CoreAudioOutput::Close()
{
	aout_sys_t   *p_sys = _sys;
    OSStatus            err = noErr;
    UInt32              i_param_size = 0;
    
    cerr << "in  CLOSE" << endl;

    if( p_sys->au_unit )
    {
        verify_noerr( AudioOutputUnitStop( p_sys->au_unit ) );
        verify_noerr( AudioUnitUninitialize( p_sys->au_unit ) );
        verify_noerr( CloseComponent( p_sys->au_unit ) );
    }
    
    if( p_sys->b_digital )
    {
        /* Stop device */
        err = AudioDeviceStop( p_sys->i_selected_dev, 
                               (AudioDeviceIOProc)RenderCallbackSPDIF ); 
        if( err != noErr )
        {
            fprintf(stderr, "AudioDeviceStop failed: [%4.4s]\n", (char *)&err );
        }

        /* Remove IOProc callback */
        err = AudioDeviceRemoveIOProc( p_sys->i_selected_dev,
                                       (AudioDeviceIOProc)RenderCallbackSPDIF );
        if( err != noErr )
        {
            fprintf(stderr, "AudioDeviceRemoveIOProc failed: [%4.4s]\n", (char *)&err );
        }
	cerr << "closing!!!!!!! " << endl;

        if( p_sys->b_revert )
        {
            AudioStreamChangeFormat(p_sys->i_stream_id, p_sys->sfmt_revert );
        }

        if( p_sys->b_changed_mixing && p_sys->sfmt_revert.mFormatID != kAudioFormat60958AC3 )
        {
            int b_mix;
            Boolean b_writeable;
            /* Revert mixable to true if we are allowed to */
            err = AudioDeviceGetPropertyInfo( p_sys->i_selected_dev, 0, FALSE, kAudioDevicePropertySupportsMixing,
                                        &i_param_size, &b_writeable );

            err = AudioDeviceGetProperty( p_sys->i_selected_dev, 0, FALSE, kAudioDevicePropertySupportsMixing,
                                        &i_param_size, &b_mix );
                                        
            if( !err && b_writeable )
            {
                fprintf(stderr, "mixable is: %d\n", b_mix );
                b_mix = 1;
                err = AudioDeviceSetProperty( p_sys->i_selected_dev, 0, 0, FALSE,
                                    kAudioDevicePropertySupportsMixing, i_param_size, &b_mix );
            }

            if( err != noErr )
            {
                fprintf(stderr, "failed to set mixmode: [%4.4s]\n", (char *)&err );
            }
        }
    }

    err = AudioHardwareRemovePropertyListener( kAudioHardwarePropertyDevices,
                                               HardwareListener );
                                               
    if( err != noErr )
    {
        fprintf(stderr, "AudioHardwareRemovePropertyListener failed: [%4.4s]\n", (char *)&err );
    }
    
    if( p_sys->i_hog_pid == getpid() )
    {
        p_sys->i_hog_pid = -1;
        i_param_size = sizeof( p_sys->i_hog_pid );
        err = AudioDeviceSetProperty( p_sys->i_selected_dev, 0, 0, FALSE,
                                         kAudioDevicePropertyHogMode, i_param_size, &p_sys->i_hog_pid );
        if( err != noErr ) fprintf(stderr, "Could not release hogmode: [%4.4s]\n", (char *)&err );
    }
    
    
    _opened = false;
}

OSStatus CoreAudioOutput::get_total_channels(AudioDeviceID device, int* channelCount, bool isInput) 
{
    OSStatus			err = noErr;
    UInt32				outSize;
    Boolean				outWritable;
    AudioBufferList*	bufferList = 0;
	AudioStreamID*		streamList = 0;
    int					i, numStream;
	
	err = AudioDeviceGetPropertyInfo(device, 0, isInput, kAudioDevicePropertyStreams, &outSize, &outWritable);
	if (err == noErr) {
		streamList = (AudioStreamID*)malloc(outSize);
		numStream = outSize/sizeof(AudioStreamID);
		//JCALog("get_total_channels device stream number = %ld numStream = %ld\n", device, numStream);
		err = AudioDeviceGetProperty(device, 0, isInput, kAudioDevicePropertyStreams, &outSize, streamList);
		if (err == noErr) {
			AudioStreamBasicDescription streamDesc;
			outSize = sizeof(AudioStreamBasicDescription);
			for (i = 0; i < numStream; i++) {
				err = AudioStreamGetProperty(streamList[i], 0, kAudioDevicePropertyStreamFormat, &outSize, &streamDesc);
				//JCALog("get_total_channels streamDesc mFormatFlags = %ld mChannelsPerFrame = %ld\n", streamDesc.mFormatFlags, streamDesc.mChannelsPerFrame);
			}
		}
	}
	
    *channelCount = 0;
    err = AudioDeviceGetPropertyInfo(device, 0, isInput, kAudioDevicePropertyStreamConfiguration, &outSize, &outWritable);
    if (err == noErr) {
        bufferList = (AudioBufferList*)malloc(outSize);
        err = AudioDeviceGetProperty(device, 0, isInput, kAudioDevicePropertyStreamConfiguration, &outSize, bufferList);
        if (err == noErr) {								
            for (i = 0; i < bufferList->mNumberBuffers; i++) 
                *channelCount += bufferList->mBuffers[i].mNumberChannels;
        }
    }
	
	if (streamList) 
		free(streamList);
	if (bufferList) 
		free(bufferList);	
		
    return err;
}


/*****************************************************************************
 * Probe: Check which devices the OS has, and add them to our audio-device menu
 *****************************************************************************/
void CoreAudioOutput::Probe()
{
    OSStatus            err = noErr;
    UInt32              i = 0, i_param_size = 0;
    AudioDeviceID       devid_def = 0;
    AudioDeviceID       *p_devices = NULL;

    aout_sys_t   *p_sys = _sys;

    /* Get number of devices */
    err = AudioHardwareGetPropertyInfo( kAudioHardwarePropertyDevices,
                                        &i_param_size, NULL );
    if( err != noErr )
    {
        fprintf(stderr, "Could not get number of devices: [%4.4s]\n", (char *)&err );
        goto error;
    }

    p_sys->i_devices = i_param_size / sizeof( AudioDeviceID );

    if( p_sys->i_devices < 1 )
    {
        fprintf(stderr, "No audio output devices were found." );
        goto error;
    }

    fprintf(stderr, "system has [%ld] device(s)\n", p_sys->i_devices );

    /* Allocate DeviceID array */
    p_devices = (AudioDeviceID*)malloc( sizeof(AudioDeviceID) * p_sys->i_devices );
    if( p_devices == NULL )
    {
        fprintf(stderr, "out of memory" );
        goto error;
    }

    /* Populate DeviceID array */
    err = AudioHardwareGetProperty( kAudioHardwarePropertyDevices,
                                    &i_param_size, p_devices );
    if( err != noErr )
    {
        fprintf(stderr, "could not get the device IDs: [%4.4s]\n", (char *)&err );
        goto error;
    }

    /* Find the ID of the default Device */
    i_param_size = sizeof( AudioDeviceID );
    err = AudioHardwareGetProperty( kAudioHardwarePropertyDefaultOutputDevice,
                                    &i_param_size, &devid_def );
    if( err != noErr )
    {
        fprintf(stderr, "could not get default audio device: [%4.4s]\n", (char *)&err );
        goto error;
    }
    p_sys->i_default_dev = devid_def;
    
    _device_list.clear();

    for( i = 0; i < p_sys->i_devices; i++ )
    {
        char *psz_name;
        i_param_size = 0;

        /* Retrieve the length of the device name */
        err = AudioDeviceGetPropertyInfo(
                    p_devices[i], 0, VLC_FALSE,
                    kAudioDevicePropertyDeviceName,
                    &i_param_size, NULL);
        if( err ) goto error;

        /* Retrieve the name of the device */
        psz_name = (char *)malloc( i_param_size );
        err = AudioDeviceGetProperty(
                    p_devices[i], 0, VLC_FALSE,
                    kAudioDevicePropertyDeviceName,
                    &i_param_size, psz_name);
        if( err ) goto error;

        fprintf(stderr, "DevID: %#lx DevName: %s\n", p_devices[i], psz_name );

        if( !AudioDeviceHasOutput( p_devices[i]) )
        {
            fprintf(stderr, "this device is INPUT only. skipping...\n" );
            continue;
        }

        /* Add the menu entries */
	//_device_list.push_back(DeviceInfo((const char *) psz_name, (int) p_devices[i]));
	
        if( p_sys->i_default_dev == p_devices[i] )
        {
        }
	
	// for now only accept devices that do digital out (or override with allow_pcm)
	// this populates our device list
        AudioDeviceSupportsDigital(p_devices[i], psz_name);


        free( psz_name);
    }
    
    for (DeviceInfoList::iterator iter = _device_list.begin(); iter != _device_list.end(); ++iter) {
	    cerr << "Device:  " << iter->first  << " = " << iter->second << endl;
    }

    /* If we change the device we want to use, we should renegotiate the audio chain */
    // var_AddCallback( p_aout, "audio-device", AudioDeviceCallback, NULL );

    /* Attach a Listener so that we are notified of a change in the Device setup */
    //err = AudioHardwareAddPropertyListener( kAudioHardwarePropertyDevices,
    //                                        HardwareListener, 
    //
    //                                    (void *)this );
    if( err )
        goto error;

    if( p_devices ) free( p_devices );
    return;

error:
    if( p_devices ) free( p_devices );
    return;
}

/*****************************************************************************
 * AudioDeviceHasOutput: Checks if the Device actually provides any outputs at all
 *****************************************************************************/
int CoreAudioOutput::AudioDeviceHasOutput( AudioDeviceID i_dev_id )
{
    UInt32			dataSize;
    Boolean			isWritable;
	
    verify_noerr( AudioDeviceGetPropertyInfo( i_dev_id, 0, FALSE, kAudioDevicePropertyStreams, &dataSize, &isWritable) );
    if (dataSize == 0) return FALSE;
    
    return TRUE;
}

/*****************************************************************************
 * AudioDeviceSupportsDigital: Check i_dev_id for digital stream support.
 *****************************************************************************/
int CoreAudioOutput::AudioDeviceSupportsDigital(AudioDeviceID i_dev_id, const char * devname )
{
    OSStatus                    err = noErr;
    UInt32                      i_param_size = 0;
    AudioStreamID               *p_streams = NULL;
    int                         i = 0, i_streams = 0;
    int                         b_return = VLC_FALSE;
    
    /* Retrieve all the output streams */
    err = AudioDeviceGetPropertyInfo( i_dev_id, 0, FALSE,
                                      kAudioDevicePropertyStreams,
                                      &i_param_size, NULL );
    if( err != noErr )
    {
        fprintf(stderr, "could not get number of streams: [%4.4s]\n", (char *)&err );
        return VLC_FALSE;
    }
    
    i_streams = i_param_size / sizeof( AudioStreamID );
    p_streams = (AudioStreamID *)malloc( i_param_size );
    if( p_streams == NULL )
    {
        fprintf(stderr, "Out of memory" );
        return -2;
    }
    
    err = AudioDeviceGetProperty( i_dev_id, 0, FALSE,
                                    kAudioDevicePropertyStreams,
                                    &i_param_size, p_streams );
    
    if( err != noErr )
    {
        fprintf(stderr, "could not get number of streams: [%4.4s]\n", (char *)&err );
        return VLC_FALSE;
    }

    for( i = 0; i < i_streams; i++ )
    {
	    if( (err = AudioStreamSupportsDigital(i_dev_id, devname, p_streams[i], i )) > 0 )
	    {
		    b_return = VLC_TRUE;
		    if (err > 1) {
			    // skip the next stream
			    ++i;
		    }
	    }
    }
    
    if( p_streams ) free( p_streams );
    return b_return;
}

/*****************************************************************************
 * AudioStreamSupportsDigital: Check i_stream_id for digital stream support.
 *****************************************************************************/
int CoreAudioOutput::AudioStreamSupportsDigital(AudioDeviceID i_device_id, const char * devname, AudioStreamID i_stream_id, int i_stream_index )
{
    OSStatus                    err = noErr;
    UInt32                      i_param_size = 0;
    AudioStreamBasicDescription *p_format_list = NULL;
    int                         i = 0, i_formats = 0;
    int                  b_return = VLC_FALSE;
    char                 tmpstr[30];
    char                 namestr[80];
    
    /* Retrieve all the stream formats supported by each output stream */
    err = AudioStreamGetPropertyInfo( i_stream_id, 0,
                                      kAudioStreamPropertyPhysicalFormats,
                                      &i_param_size, NULL );
    if( err != noErr )
    {
        fprintf(stderr, "could not get number of streamformats: [%4.4s]\n", (char *)&err );
        return VLC_FALSE;
    }
    
    i_formats = i_param_size / sizeof( AudioStreamBasicDescription );
    p_format_list = (AudioStreamBasicDescription *)malloc( i_param_size );
    if( p_format_list == NULL )
    {
        fprintf(stderr, "could not malloc the memory" );
        return VLC_FALSE;
    }
    
    err = AudioStreamGetProperty( i_stream_id, 0,
                                      kAudioStreamPropertyPhysicalFormats,
                                      &i_param_size, p_format_list );
    if( err != noErr )
    {
        fprintf(stderr, "could not get the list of streamformats: [%4.4s]\n", (char *)&err );
        free( p_format_list);
        p_format_list = NULL;
        return VLC_FALSE;
    }

    for( i = 0; i < i_formats; i++ )
    {
        fprintf(stderr, STREAM_FORMAT_MSG( "supported format: ", p_format_list[i] ) );
	fprintf(stderr, "\n");        
        if( (p_format_list[i].mFormatID == 'IAC3' ||
	     p_format_list[i].mFormatID == kAudioFormat60958AC3) && p_format_list[i].mSampleRate == _samplerate )
        {
		snprintf(tmpstr, sizeof(tmpstr), "%d", (int) i_device_id | AOUT_VAR_SPDIF_FLAG);
		_device_list.push_back(DeviceInfo(string(devname) + string(" (AC3)"), string(tmpstr)));

		b_return = 1;
        }
    }
    
    if (_allow_pcm) 
    {
	    // fill in options for PCM streams too
	    for( i = 0; i < i_formats; i++ )
	    {
		    if (p_format_list[i].mFormatID == kAudioFormatLinearPCM
			&& (p_format_list[i].mFormatFlags & kAudioFormatFlagIsSignedInteger ))
		    {
			    if (p_format_list[i].mChannelsPerFrame == 1) {
				    // skip next stream, we'll use it as the 
				    // second
				    b_return = 2;
				    snprintf(namestr, sizeof(namestr), "%s:%d+%d", devname,i_stream_index+1, i_stream_index+2);
			    }
			    else {
				    snprintf(namestr, sizeof(namestr), "%s", devname);
				    b_return = 1;
	    }
			    snprintf(tmpstr, sizeof(tmpstr), "%d:%d", (int) i_device_id, i_stream_index);
			    _device_list.push_back(DeviceInfo(string(namestr), string(tmpstr)));
			    break;
		    }
	    }
    }

    if( p_format_list ) free( p_format_list );
    return b_return;
}

/*****************************************************************************
 * AudioStreamChangeFormat: Change i_stream_id to change_format
 *****************************************************************************/
int CoreAudioOutput::AudioStreamChangeFormat( AudioStreamID i_stream_id, AudioStreamBasicDescription change_format )
{
    OSStatus            err = noErr;
    UInt32              i_param_size = 0;
    int i;
    int ret = VLC_TRUE;

    struct timeval now;
    struct timespec timeout;
    struct { pthread_mutex_t lock; pthread_cond_t cond; } w;
    
    fprintf(stderr, STREAM_FORMAT_MSG( "setting stream format: ", change_format ) );
    fprintf(stderr, "\n");
    /* Condition because SetProperty is asynchronious */ 
    // Lock

    pthread_cond_init(&w.cond, NULL);
    pthread_mutex_init( &w.lock, NULL );
    pthread_mutex_lock( &w.lock );

    /* Install the callback */
    err = AudioStreamAddPropertyListener( i_stream_id, 0,
                                     kAudioStreamPropertyPhysicalFormat,
                                     StreamListener, (void *)&w );
    if( err != noErr )
    {
       fprintf(stderr, "AudioStreamAddPropertyListener failed: [%4.4s]", (char *)&err );
	ret = VLC_FALSE;
	goto cleanup;
    }

    /* change the format */
    err = AudioStreamSetProperty( i_stream_id, 0, 0,
                                  kAudioStreamPropertyPhysicalFormat,
                                  sizeof( AudioStreamBasicDescription ),
                                  &change_format ); 
    if( err != noErr )
    {
        fprintf(stderr, "could not set the stream format: [%4.4s]\n", (char *)&err );
	ret = VLC_FALSE;
	goto cleanup;
    }

    /* The AudioStreamSetProperty is not only asynchronious (requiring the locks)
     * it is also not Atomic, in it's behaviour.
     * Therefore we check 5 times before we really give up.
     * FIXME: failing isn't actually implemented yet. */
    for( i = 0; i < 5; i++ )
    {
        AudioStreamBasicDescription actual_format;

        gettimeofday( &now, NULL );
        timeout.tv_sec = now.tv_sec;
        timeout.tv_nsec = (now.tv_usec + 500000) * 1000;

        if( pthread_cond_timedwait( &w.cond, &w.lock, &timeout ) )
        {
	   fprintf(stderr, "reached timeout" );
        }

        i_param_size = sizeof( AudioStreamBasicDescription );
        err = AudioStreamGetProperty( i_stream_id, 0,
                                      kAudioStreamPropertyPhysicalFormat,
                                      &i_param_size, 
                                      &actual_format );

        fprintf(stderr, STREAM_FORMAT_MSG( "actual format in use: ", actual_format ) );
	fprintf(stderr, "\n");
        if( actual_format.mSampleRate == change_format.mSampleRate &&
            actual_format.mFormatID == change_format.mFormatID &&
            actual_format.mFramesPerPacket == change_format.mFramesPerPacket )
        {
            /* The right format is now active */
            break;
        }
        /* We need to check again */
    }
    
    /* Removing the property listener */
    err = AudioStreamRemovePropertyListener( i_stream_id, 0,
                                           kAudioStreamPropertyPhysicalFormat,
                                           StreamListener );
    if( err != noErr )
    {
       fprintf(stderr, "AudioStreamRemovePropertyListener failed: [%4.4s]", (char *)&err );
       ret = VLC_FALSE;
       goto cleanup;
    }
    
    /* Destroy the lock and condition */
cleanup: 
    pthread_mutex_unlock( &w.lock );
    pthread_mutex_destroy( &w.lock );
    pthread_cond_destroy( &w.cond );
    
    return ret;
}

/*****************************************************************************
 * RenderCallbackAnalog: This function is called everytime the AudioUnit wants
 * us to provide some more audio data.
 * Don't print anything during normal playback, calling blocking function from
 * this callback is not allowed.
 *****************************************************************************/
OSStatus CoreAudioOutput::RenderCallbackAnalog( void *_pthis,
                                      AudioUnitRenderActionFlags *ioActionFlags,
                                      const AudioTimeStamp *inTimeStamp,
                                      unsigned int inBusNummer,
                                      unsigned int inNumberFrames,
                                      AudioBufferList *ioData )
{
	//AudioTimeStamp  host_time;
	//double         current_date = 0;
	//uint32_t        i_mData_bytes = 0;    

    CoreAudioOutput * pthis = (CoreAudioOutput *) _pthis;
    aout_sys_t * p_sys = pthis->_sys;

    //host_time.mFlags = kAudioTimeStampHostTimeValid;
    //AudioDeviceTranslateTime( p_sys->i_selected_dev, inTimeStamp, &host_time );

    /* Check for the difference between the Device clock and mdate */
    //p_sys->clock_diff = - (double)
    //   AudioConvertHostTimeToNanos( AudioGetCurrentHostTime() ) / 1000; 
    //p_sys->clock_diff += mdate();

    //current_date = p_sys->clock_diff +
    //              AudioConvertHostTimeToNanos( host_time.mHostTime ) / 1000;
                   //- ((double) 1000000 / p_aout->output.output.i_rate * 31 ); // 31 = Latency in Frames. retrieve somewhere

    if( ioData == NULL && ioData->mNumberBuffers < 1 )
    {
        fprintf(stderr, "no iodata or buffers\n");
        return 0;
    }
    if( ioData->mNumberBuffers > 1 )
        fprintf(stderr, "well this is weird. seems like there is more than one buffer...\n" );


    // callback to copy data into buffers
#define BUFFER ioData->mBuffers[0]
    
    uint32_t retbytes = pthis->_source->get_output((unsigned char *) BUFFER.mData,
						   BUFFER.mDataByteSize);
    //cerr << " mchannels is " << BUFFER.mNumberChannels << "  ioData, databyte size is: " << BUFFER.mDataByteSize << " retbytes: " << retbytes << endl;

    if (retbytes < BUFFER.mDataByteSize) {
	    // silence the remainder
	    memset((unsigned char *) BUFFER.mData + retbytes, 0, BUFFER.mDataByteSize - retbytes);
	    //cerr << " and " << outBytes-towrite << "  silence" << endl;
    }
#undef BUFFER
   
    return( noErr );     
}

/*****************************************************************************
 * RenderCallbackSPDIF: callback for SPDIF audio output
 *****************************************************************************/
OSStatus CoreAudioOutput::RenderCallbackSPDIF( AudioDeviceID inDevice,
                                    const AudioTimeStamp * inNow, 
                                    const void * inInputData,
                                    const AudioTimeStamp * inInputTime, 
                                    AudioBufferList * outOutputData,
                                    const AudioTimeStamp * inOutputTime, 
                                    void * threadGlobals )
{
	//aout_buffer_t * p_buffer;
	//double         current_date;

    CoreAudioOutput * pthis = (CoreAudioOutput *) threadGlobals;
    aout_sys_t * p_sys = pthis->_sys;

    /* Check for the difference between the Device clock and mdate */
    //p_sys->clock_diff = - (double)
    //    AudioConvertHostTimeToNanos( inNow->mHostTime ) / 1000; 
    //p_sys->clock_diff += mdate();

    //current_date = p_sys->clock_diff +
    //              AudioConvertHostTimeToNanos( inOutputTime->mHostTime ) / 1000;
                   //- ((double) 1000000 / p_aout->output.output.i_rate * 31 ); // 31 = Latency in Frames. retrieve somewhere

    // callback to copy data

    if (p_sys->stream_format.mChannelsPerFrame == 2) {
            #define BUFFER outOutputData->mBuffers[p_sys->i_stream_index]
	    
	    uint32_t retbytes = pthis->_source->get_output((unsigned char *) BUFFER.mData,
							   BUFFER.mDataByteSize);
	    //cerr << " mchannels is 2, databyte size is: " << BUFFER.mDataByteSize << endl;
	    if (retbytes < BUFFER.mDataByteSize) {
		    // silence the remainder
		    memset((unsigned char *) BUFFER.mData + retbytes, 0, BUFFER.mDataByteSize - retbytes);
		    //cerr << " and " << BUFFER.mDataByteSize-retbytes << "  silence" << endl;
	    }

           #undef BUFFER
    }
    else if (p_sys->stream_format.mChannelsPerFrame == 1) {
	    // THIS CRAP IS NOT REALLY NECESSARY AND DOESN"T REALLY WORK
            #define BUFFER outOutputData->mBuffers[p_sys->i_stream_index]
            #define BUFFER2 outOutputData->mBuffers[p_sys->i_stream_index + 1]

	    //cerr << " mchannels is 1, deinte databyte size is: " << BUFFER.mDataByteSize << "  byteperpack: " << p_sys->stream_format.mBytesPerPacket
	    // << "  framperpack: " << p_sys->stream_format.mFramesPerPacket << endl;

	    // we must be using two streams and deinterleave
	    uint32_t retbytes = pthis->_source->get_output ((unsigned char *) BUFFER.mData,(unsigned char *) BUFFER2.mData,
							   BUFFER.mDataByteSize);
	    
	    if (retbytes < BUFFER.mDataByteSize) {
		    // silence the remainder
		    memset((unsigned char *) BUFFER.mData + retbytes, 0, BUFFER.mDataByteSize - retbytes);
		    memset((unsigned char *) BUFFER2.mData + retbytes, 0, BUFFER.mDataByteSize - retbytes);
		    //cerr << " and " << outBytes-towrite << "  silence" << endl;
	    }
	    #undef BUFFER
	    #undef BUFFER2
    }

    return( noErr );     
}

/*****************************************************************************
 * HardwareListener: Warns us of changes in the list of registered devices
 *****************************************************************************/
OSStatus CoreAudioOutput::HardwareListener( AudioHardwarePropertyID inPropertyID,
                                  void * inClientData )
{
    OSStatus err = noErr;
    CoreAudioOutput     *pthis = (CoreAudioOutput *)inClientData;

    switch( inPropertyID )
    {
        case kAudioHardwarePropertyDevices:
        {
            /* something changed in the list of devices */
            /* We trigger the audio-device's aout_ChannelsRestart callback */
		// var_Change( p_aout, "audio-device", VLC_VAR_TRIGGER_CALLBACKS, NULL, NULL );
		//var_Destroy( p_aout, "audio-device" );
        }
        break;
    }

    return( err );
}

/*****************************************************************************
 * StreamListener 
 *****************************************************************************/
OSStatus CoreAudioOutput::StreamListener( AudioStreamID inStream,
                                UInt32 inChannel,
                                AudioDevicePropertyID inPropertyID,
                                void * inClientData )
{
    OSStatus err = noErr;
    typedef struct { pthread_mutex_t lock; pthread_cond_t cond; } LockCond;
    LockCond* w = (LockCond*) inClientData;
    
    switch( inPropertyID )
    {
        case kAudioStreamPropertyPhysicalFormat:
            pthread_mutex_lock( &w->lock );
            pthread_cond_signal( &w->cond );
            pthread_mutex_unlock( &w->lock ); 
            break;

        default:
            break;
    }

    return( err );
}

#if 0
/*****************************************************************************
 * AudioDeviceCallback: Callback triggered when the audio-device variable is changed
 *****************************************************************************/
static int CoreAudioOutput::AudioDeviceCallback( const char *psz_variable,
                     vlc_value_t old_val, vlc_value_t new_val, void *param )
{
	//aout_instance_t *p_aout = (aout_instance_t *)p_this;
	//var_Set( p_aout->p_vlc, "macosx-audio-device", new_val );
	//fprintf(stderr, "Set Device: %#x", new_val.i_int );
	//return aout_ChannelsRestart( p_this, psz_variable, old_val, new_val, param );
	return 0;
}
#endif

void CoreAudioOutput::GetDevices(DeviceInfoList & retlist)
{
	cerr << "GET DEVICES: " << endl;
	Probe();

	retlist.insert(retlist.begin(), _device_list.begin(), _device_list.end());
}

void CoreAudioOutput::set_allow_pcm (bool val)
{
	if (_allow_pcm != val) {
		_allow_pcm = val;
		cerr << "set allow pcm to: " << val << endl;
		Probe();
	}
}
