/*
 *
 */

#include <CoreAudio/CoreAudio.h>
#include <AudioUnit/AudioUnitProperties.h>
#include <AudioUnit/AudioUnitParameters.h>
#include <AudioUnit/AudioOutputUnit.h>
#include <AudioToolbox/AudioFormat.h>

#include <string>
#include <list>

#include "types.hpp"

namespace AC3JACK {
namespace COREAUDIO {

// for callbacks to get actual data to output
class DataSource {
public:
	virtual ~DataSource() {}
	// interleaved data
	virtual uint32_t get_output(unsigned char * outdata, uint32_t databytes) = 0;
	// deinterleave data into two buffers
	virtual uint32_t get_output(unsigned char * outdata, unsigned char * outdata2, uint32_t databytes) = 0;

};


class CoreAudioOutput
{
public:

	CoreAudioOutput(DataSource * source, double samplerate);
	virtual ~CoreAudioOutput();

	void set_allow_pcm (bool val);
	bool get_allow_pcm() const { return _allow_pcm; }

	void set_device_stream (const std::string & idinfo);

	int      Open                    ();
	void     Close                   ();
	
	void     Probe                   ();
	void     GetDevices (DeviceInfoList & retlist);

protected:
	int      OpenAnalog              ();
	int      OpenSPDIF               ();
	
	int      AudioDeviceHasOutput    (AudioDeviceID );
	int      AudioDeviceSupportsDigital (AudioDeviceID, const char * devname );
	int      AudioStreamSupportsDigital (AudioDeviceID,  const char * devname, AudioStreamID i_stream_id, int i_stream_index);
	int      AudioStreamChangeFormat (AudioStreamID, AudioStreamBasicDescription );

	OSStatus get_total_channels(AudioDeviceID device, int* channelCount, bool isInput=false);
	
	static OSStatus RenderCallbackAnalog    ( void *, AudioUnitRenderActionFlags *, const AudioTimeStamp *,
                                          unsigned int, unsigned int, AudioBufferList *);
	static OSStatus RenderCallbackSPDIF     ( AudioDeviceID, const AudioTimeStamp *, const void *, const AudioTimeStamp *,
						  AudioBufferList *, const AudioTimeStamp *, void * );
	static OSStatus HardwareListener        ( AudioHardwarePropertyID, void *);
	static OSStatus StreamListener          ( AudioStreamID, UInt32,
						  AudioDevicePropertyID, void * );
	//static int      AudioDeviceCallback     ( void *, const char *,
        //                                  vlc_value_t, vlc_value_t, void * );



protected:
	
/*****************************************************************************
 * aout_sys_t: private audio output method descriptor
 *****************************************************************************
 * This structure is part of the audio output thread descriptor.
 * It describes the CoreAudio specific properties of an output thread.
 *****************************************************************************/
	typedef struct
	{
		AudioDeviceID               i_default_dev;  /* Keeps DeviceID of defaultOutputDevice */
		AudioDeviceID               i_selected_dev; /* Keeps DeviceID of the selected device */
		UInt32                      i_devices;      /* Number of CoreAudio Devices */
		bool                  b_supports_digital;/* Does the currently selected device support digital mode? */
		bool                  b_digital;      /* Are we running in digital mode? */
		double                     clock_diff;     /* Difference between VLC clock and Device clock */
		
		/* AUHAL specific */
		Component                   au_component;   /* The Audiocomponent we use */
		AudioUnit                   au_unit;        /* The AudioUnit we use */
		//uint8_t                     p_remainder_buffer[BUFSIZE];
		uint32_t                    i_read_bytes;
		uint32_t                    i_total_bytes;
		
		/* CoreAudio SPDIF mode specific */
		pid_t                       i_hog_pid;      /* The keep the pid of our hog status */
		AudioStreamID               i_stream_id;    /* The StreamID that has a cac3 streamformat */
		int                         i_stream_index; /* The index of i_stream_id in an AudioBufferList */
		AudioStreamBasicDescription stream_format;  /* The format we changed the stream to */
		AudioStreamBasicDescription sfmt_revert;    /* The original format of the stream */
		bool                  b_revert;       /* Wether we need to revert the stream format */
		bool                  b_changed_mixing;/* Wether we need to set the mixing mode back */
	} aout_sys_t;
	
	aout_sys_t    *_sys;
	std::string    _devstr;
	double        _samplerate;
	double        _output_rate;
	bool          _opened;
	bool          _hw_bigendian;
	bool          _allow_pcm;

	DeviceInfoList _device_list;
	DataSource     * _source;
};
}
}
