/*
** Copyright (C) 2006 Jesse Chappell <jesse@essej.net>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/

#include <iostream>
#include "boost/date_time/posix_time/posix_time.hpp"
#include "boost/thread/thread.hpp"
#include "boost/thread/xtime.hpp"

#include <cstdio>
#include <cerrno>
#include <assert.h>
#include <alloca.h>
#include <cstring>
#include <unistd.h>

#include "coreaudio_spdif_writer.hpp"

using namespace AC3JACK;
using namespace std;
using namespace boost;

#define checkErr( err) \
if(err) {\
	OSStatus error = static_cast<OSStatus>(err);\
		fprintf(stderr, "Error: %ld ->  %s: %d\n",  error,\
			   __FILE__, \
			   __LINE__\
			   );\
				   fflush(stderr);\
		return err; \
}         

CoreAudioSpdifWriter::CoreAudioSpdifWriter(unsigned int samplerate, const std::string & devstr, size_t bufsize)
	: SpdifWriter(), _devstr(devstr), _samplerate(samplerate), _ca_out(0)
{
	_buffer = new RingBuffer<unsigned char>(bufsize);
	_ok = false;

	_ca_out = new COREAUDIO::CoreAudioOutput(this, (double) _samplerate);
	_ca_out->Probe();
	
	DeviceInfoList infolist;
	_ca_out->GetDevices(infolist);
	//if (!infolist.empty()) {
	//	_ca_out->set_device_stream (infolist.front().second);
	//}
}

void CoreAudioSpdifWriter::start()
{
	if (_ca_out->Open() == 1) {
		_ok = true;
	}
	else {
		_status = "Error: could not open any device for SPDIF output";
	}
}

CoreAudioSpdifWriter::~CoreAudioSpdifWriter()
{
	if (_ca_out) {
		delete _ca_out;
	}

	delete _buffer;
}






size_t CoreAudioSpdifWriter::write(unsigned char * data, size_t length)
{
	size_t wrote;
	//cerr << "writing " << length << " into buffer" << endl;
	wrote = _buffer->write (data, length);

	return wrote;
}


uint32_t CoreAudioSpdifWriter::get_output(unsigned char * outdata,
					  uint32_t outbytes)
{
	// get the data from ring buffer
	// ioData, inNumberFrames
	uint32_t towrite = 0;

	towrite = min(outbytes, (uint32_t) _buffer->read_space());

	if (towrite > 0) {
		towrite = _buffer->read (outdata, towrite);
		// cerr << "CA spdif: wrote " << towrite << " bytes" << endl;
	}

	return towrite;

	// silence the remainder
	//if (outbytes > towrite) {
	//	memset(outBuffer + towrite, 0, outBytes - towrite);
		//cerr << " and " << outBytes-towrite << "  silence" << endl;
	//}
}

uint32_t CoreAudioSpdifWriter::get_output(unsigned char * outdata, unsigned char * outdata2, uint32_t databytes)
{
	// get the data from ring buffer
	// ioData, inNumberFrames
	uint32_t towrite = 0;
	RingBuffer<unsigned char>::rw_vector  rvec;
	int * obuf1 = (int *) outdata;
	int * obuf2 = (int *) outdata2;
	short * sbuf; 
	uint32_t retlen = 0;

	//cerr << "get output deinterleave" << endl;

	_buffer->get_read_vector(&rvec);
	sbuf = (short *) rvec.buf[0];

	databytes /= 2; // they want databytes per buffer and they are using 32 bit samples

	towrite = min(databytes, (uint32_t) rvec.len[0]);

	if (towrite > 0) {
		// deinterleave 16 bit chunks (TODO: optimize) into 24-in-32
		for (uint32_t i=0; i < towrite/2; i+=2) {
			*obuf1++ = ((int)sbuf[i] << 16);
			*obuf2++ =  ((int)sbuf[i+1] << 16);
		}
		databytes -= towrite;
		retlen = towrite;

		towrite = min(databytes, (uint32_t) rvec.len[1]);
		
		if (towrite > 0) {
			sbuf = (short *) rvec.buf[1];

			for (uint32_t i=0; i < towrite/2; i+=2) {
				*obuf1++ = ((int)sbuf[i] << 16);
				*obuf2++ =  ((int)sbuf[i+1] << 16);
			}
			retlen += towrite;
		}

		//cerr << "CA deinter spdif: wrote " << retlen*2 << " bytes to each" << endl;
		_buffer->increment_read_ptr(retlen);
	} else {
		//cerr << " to write is only: " << towrite << "  rvec.len[0] = " << rvec.len[0] << "  readps: " << _buffer->read_space() << endl;
	}

	return retlen*2;
}


int CoreAudioSpdifWriter::get_devices(DeviceInfoList & retlist)
{
	if (_ca_out) {
		_ca_out->GetDevices(retlist);
		return retlist.size();
	}
	return 0;
}


void CoreAudioSpdifWriter::set_allow_pcm (bool val)
{
	if (_ca_out) {
		_ca_out->set_allow_pcm(val);
	}
}

void CoreAudioSpdifWriter::set_device (std::string devstr)
{
	if (_devstr != devstr) {
		cerr << "Set dives to " << devstr << endl;
		_devstr = devstr;
		if (_ca_out) {
			_ca_out->set_device_stream(_devstr);
		}	
	}
}
